from typing import Any, Self, Type, TypeVar

from pydantic import BaseModel
from sqlalchemy.orm import DeclarativeBase

IDENTIFIABLE_SCHEMA_T = TypeVar("IDENTIFIABLE_SCHEMA_T")


T_BASEMODEL = TypeVar("T_BASEMODEL", bound=BaseModel)


def recursive_get_dict(obj: Any) -> Any:
    if isinstance(obj, list):
        return [recursive_get_dict(v) for v in obj]
    elif hasattr(obj, "__dict__"):
        return {
            k: recursive_get_dict(v) for k, v in obj.__dict__.items() if k[0] != "_"
        }
    else:
        return obj


class BaseEntity(DeclarativeBase):

    @classmethod
    def from_basemodel(cls, mutation: T_BASEMODEL) -> "Self":
        intersection = set(cls.__annotations__.keys()) & set(
            mutation.model_fields.keys()
        )
        return cls(**{k: getattr(mutation, k) for k in intersection})

    def to_basemodel(self, model: Type[T_BASEMODEL]) -> T_BASEMODEL:
        return model.model_validate(recursive_get_dict(self))



import inspect
from dataclasses import dataclass
from typing import Any, Awaitable, Callable, Generic, TypeVar, cast

from jararaca.messagebus.types import INHERITS_MESSAGE_CO, Message, MessageOf

DECORATED_FUNC = TypeVar("DECORATED_FUNC", bound=Callable[..., Any])
DECORATED_CLASS = TypeVar("DECORATED_CLASS", bound=Any)


class MessageHandler(Generic[INHERITS_MESSAGE_CO]):

    MESSAGE_INCOMING_ATTR = "__message_incoming__"

    def __init__(
        self,
        message: type[INHERITS_MESSAGE_CO],
        timeout: int | None = None,
        exception_handler: Callable[[BaseException], None] | None = None,
        nack_on_exception: bool = False,
        auto_ack: bool = True,
    ) -> None:
        self.message_type = message

        self.timeout = timeout
        self.exception_handler = exception_handler
        self.requeue_on_exception = nack_on_exception
        self.auto_ack = auto_ack

    def __call__(
        self, func: Callable[[Any, MessageOf[INHERITS_MESSAGE_CO]], Awaitable[None]]
    ) -> Callable[[Any, MessageOf[INHERITS_MESSAGE_CO]], Awaitable[None]]:

        MessageHandler[Any].register(func, self)

        return func

    @staticmethod
    def register(
        func: Callable[[Any, MessageOf[INHERITS_MESSAGE_CO]], Awaitable[None]],
        message_incoming: "MessageHandler[Any]",
    ) -> None:

        setattr(func, MessageHandler.MESSAGE_INCOMING_ATTR, message_incoming)

    @staticmethod
    def get_message_incoming(
        func: Callable[[MessageOf[Any]], Awaitable[Any]]
    ) -> "MessageHandler[Message] | None":
        if not hasattr(func, MessageHandler.MESSAGE_INCOMING_ATTR):
            return None

        return cast(
            MessageHandler[Message], getattr(func, MessageHandler.MESSAGE_INCOMING_ATTR)
        )


@dataclass(frozen=True)
class MessageHandlerData:
    message_type: type[Any]
    spec: MessageHandler[Message]
    callable: Callable[[MessageOf[Any]], Awaitable[None]]


MESSAGE_HANDLER_DATA_SET = set[MessageHandlerData]


class MessageBusController:

    MESSAGEBUS_ATTR = "__messagebus__"

    def __init__(self) -> None:
        self.messagebus_factory: Callable[[Any], MESSAGE_HANDLER_DATA_SET] | None = None

    def get_messagebus_factory(
        self,
    ) -> Callable[[DECORATED_CLASS], MESSAGE_HANDLER_DATA_SET]:
        if self.messagebus_factory is None:
            raise Exception("MessageBus factory is not set")
        return self.messagebus_factory

    def __call__(self, func: type[DECORATED_CLASS]) -> type[DECORATED_CLASS]:

        def messagebus_factory(
            instance: DECORATED_CLASS,
        ) -> MESSAGE_HANDLER_DATA_SET:
            handlers: MESSAGE_HANDLER_DATA_SET = set()
            inspect.signature(func)

            members = inspect.getmembers(func, predicate=inspect.isfunction)

            for name, member in members:
                message_incoming = MessageHandler.get_message_incoming(member)

                if message_incoming is None:
                    continue

                if not inspect.iscoroutinefunction(member):
                    raise Exception(
                        "Message incoming handler '%s' from '%s.%s' must be a coroutine function"
                        % (name, func.__module__, func.__qualname__)
                    )

                handlers.add(
                    MessageHandlerData(
                        message_type=message_incoming.message_type,
                        spec=message_incoming,
                        callable=getattr(instance, name),
                    )
                )

            return handlers

        self.messagebus_factory = messagebus_factory

        MessageBusController.register(func, self)

        return func

    @staticmethod
    def register(
        func: type[DECORATED_CLASS], messagebus: "MessageBusController"
    ) -> None:

        setattr(func, MessageBusController.MESSAGEBUS_ATTR, messagebus)

    @staticmethod
    def get_messagebus(func: type[DECORATED_CLASS]) -> "MessageBusController | None":
        if not hasattr(func, MessageBusController.MESSAGEBUS_ATTR):
            return None

        return cast(
            MessageBusController, getattr(func, MessageBusController.MESSAGEBUS_ATTR)
        )



import logging
from contextlib import asynccontextmanager
from typing import AsyncContextManager, AsyncGenerator, Sequence

from jararaca.di import Container
from jararaca.microservice import (
    AppInterceptor,
    Microservice,
    is_interceptor_with_lifecycle,
)

logger = logging.getLogger(__name__)


class AppLifecycle:

    def __init__(self, app: Microservice, container: Container) -> None:
        self.app = app
        self.container = container
        self.__real_interceptors: list[AppInterceptor] | None = None

    @property
    def initialized_interceptors(self) -> Sequence[AppInterceptor]:
        if self.__real_interceptors is None:
            raise ValueError("Interceptors not initialized")
        return self.__real_interceptors

    @asynccontextmanager
    async def __call__(self) -> AsyncGenerator[None, None]:

        self.__real_interceptors = []

        self.container.fill_providers(False)
        lifecycle_ctxs: list[AsyncContextManager[None]] = []

        logger.info("Initializing interceptors lifecycle")
        for interceptor_dep in self.app.interceptors:
            interceptor: AppInterceptor
            if not isinstance(interceptor_dep, AppInterceptor):
                interceptor = self.container.get_or_register_token_or_type(
                    interceptor_dep
                )

            else:
                interceptor = interceptor_dep
            self.__real_interceptors.append(interceptor)

            if not is_interceptor_with_lifecycle(interceptor):
                continue

            ctx = interceptor.lifecycle(self.app, self.container)
            lifecycle_ctxs.append(ctx)

            await ctx.__aenter__()

        self.container.fill_providers(True)

        yield

        logger.info("Finalizing interceptors lifecycle")
        for ctx in lifecycle_ctxs:
            await ctx.__aexit__(None, None, None)



import importlib
import importlib.resources
from codecs import StreamWriter
from typing import Any
from urllib.parse import urlparse, urlunsplit

import click
import uvicorn
from mako.template import Template  # type: ignore

from jararaca.messagebus.worker import AioPikaWorkerConfig, MessageBusWorker
from jararaca.microservice import Microservice
from jararaca.presentation.http_microservice import HttpMicroservice
from jararaca.presentation.server import create_http_server
from jararaca.scheduler.scheduler import Scheduler, SchedulerBackend, SchedulerConfig
from jararaca.tools.typescript.interface_parser import (
    write_microservice_to_typescript_interface,
)

LIBRARY_FILES_PATH = importlib.resources.files("jararaca.files")
ENTITY_TEMPLATE_PATH = LIBRARY_FILES_PATH / "entity.py.mako"


def find_item_by_module_path(
    module_path: str,
) -> Any:
    if ":" not in module_path:
        raise ValueError("'%s' is not a valid module path" % module_path)

    module_name, app = module_path.rsplit(":", 1)

    try:
        module = importlib.import_module(module_name)
    except ImportError as e:
        raise ImportError("App module not found") from e

    if not hasattr(module, app):
        raise ValueError("module %s has no attribute %s" % (module, app))

    app = getattr(module, app)

    return app


def find_microservice_by_module_path(module_path: str) -> Microservice:

    app = find_item_by_module_path(module_path)

    if not isinstance(app, Microservice):
        raise ValueError(
            (
                "%s must be an instance of Microservice (it is %s)"
                % (app, str(type(app)))
            )
        )

    return app


@click.group()
def cli() -> None:
    pass


@cli.command()
@click.argument(
    "app_path",
    type=str,
)
@click.option(
    "--url",
    type=str,
    default="amqp://guest:guest@localhost/",
)
@click.option(
    "--username",
    type=str,
    default=None,
)
@click.option(
    "--password",
    type=str,
    default=None,
)
@click.option(
    "--exchange",
    type=str,
    default="jararaca_ex",
)
@click.option(
    "--queue",
    type=str,
    default="jararaca_q",
)
@click.option(
    "--prefetch-count",
    type=int,
    default=1,
)
def worker(
    app_path: str,
    url: str,
    username: str | None,
    password: str | None,
    exchange: str,
    queue: str,
    prefetch_count: int,
) -> None:

    app = find_microservice_by_module_path(app_path)

    parsed_url = urlparse(url)

    if password is not None:
        parsed_url = urlparse(
            urlunsplit(
                parsed_url._replace(
                    netloc=f"{parsed_url.username or ''}:{password}@{parsed_url.netloc}"
                )
            )
        )

    if username is not None:
        parsed_url = urlparse(
            urlunsplit(
                parsed_url._replace(
                    netloc=f"{username}{':%s' % password if password is not None else ''}@{parsed_url.netloc}"
                )
            )
        )

    url = parsed_url.geturl()

    config = AioPikaWorkerConfig(
        url=url,
        exchange=exchange,
        queue=queue,
        prefetch_count=prefetch_count,
    )

    MessageBusWorker(app, config=config).start_sync()


@cli.command()
@click.argument(
    "app_path",
    type=str,
)
@click.option(
    "--host",
    type=str,
    default="0.0.0.0",
)
@click.option(
    "--port",
    type=int,
    default=8000,
)
def server(app_path: str, host: str, port: int) -> None:

    item = find_item_by_module_path(app_path)

    if isinstance(item, Microservice):
        app = HttpMicroservice(item)
    elif isinstance(item, HttpMicroservice):
        app = item
    else:
        raise ValueError(
            (
                "%s must be an instance of Microservice or HttpMicroservice (it is %s)"
                % (item, str(type(item)))
            )
        )

    asgi_app = create_http_server(app)

    uvicorn.run(asgi_app, host=host, port=port)


class NullBackend(SchedulerBackend): ...


@cli.command()
@click.argument(
    "app_path",
    type=str,
)
@click.option(
    "--interval",
    type=int,
    default=1,
)
def scheduler(
    app_path: str,
    interval: int,
) -> None:
    app = find_microservice_by_module_path(app_path)

    Scheduler(app, NullBackend(), SchedulerConfig(interval=interval)).run()


@cli.command()
@click.argument(
    "app_path",
    type=str,
)
@click.argument(
    "file_path",
    type=click.File("w"),
)
def gen_tsi(app_path: str, file_path: StreamWriter) -> None:
    app = find_microservice_by_module_path(app_path)

    content = write_microservice_to_typescript_interface(app)

    file_path.write(content)


def camel_case_to_snake_case(name: str) -> str:
    return "".join(["_" + c.lower() if c.isupper() else c for c in name]).lstrip("_")


def camel_case_to_kebab_case(name: str) -> str:
    return "".join(["-" + c.lower() if c.isupper() else c for c in name]).lstrip("-")


def camel_case_to_pascal_case(name: str) -> str:
    return name[0].upper() + name[1:]


@cli.command()
@click.argument("entity_name", type=click.STRING)
@click.argument(
    "file_path",
    type=click.File("w"),
)
def gen_entity(entity_name: str, file_path: StreamWriter) -> None:

    template = Template(filename=str(ENTITY_TEMPLATE_PATH))

    entity_snake_case = camel_case_to_snake_case(entity_name)
    entity_pascal_case = camel_case_to_pascal_case(entity_name)
    entity_kebab_case = camel_case_to_kebab_case(entity_name)

    file_path.write(
        template.render(
            entityNameSnakeCase=entity_snake_case,
            entityNamePascalCase=entity_pascal_case,
            entityNameKebabCase=entity_kebab_case,
        )
    )



from contextlib import contextmanager, suppress
from contextvars import ContextVar
from typing import Any, Generator, Protocol


class BusMessageController(Protocol):

    async def ack(self) -> None:
        pass

    async def nack(self) -> None:
        pass

    async def reject(self) -> None:
        pass

    async def retry(self) -> None:
        pass

    async def retry_later(self, delay: int) -> None:
        pass


bus_message_controller_ctxvar = ContextVar[BusMessageController](
    "bus_message_controller"
)


@contextmanager
def provide_bus_message_controller(
    controller: BusMessageController,
) -> Generator[None, Any, None]:
    token = bus_message_controller_ctxvar.set(controller)
    try:
        yield
    finally:
        with suppress(LookupError):
            bus_message_controller_ctxvar.reset(token)


def use_bus_message_controller() -> BusMessageController:
    return bus_message_controller_ctxvar.get()


async def ack() -> None:
    controller = use_bus_message_controller()
    await controller.ack()


async def nack() -> None:
    controller = use_bus_message_controller()
    await controller.nack()


async def reject() -> None:
    controller = use_bus_message_controller()
    await controller.reject()


async def retry() -> None:
    controller = use_bus_message_controller()
    await controller.retry()


async def retry_later(delay: int) -> None:
    controller = use_bus_message_controller()
    await controller.retry_later(delay)



import inspect
from contextlib import contextmanager, suppress
from contextvars import ContextVar
from dataclasses import dataclass, field
from datetime import datetime
from typing import (
    TYPE_CHECKING,
    Annotated,
    Any,
    AsyncContextManager,
    Callable,
    Generator,
    Literal,
    Protocol,
    Type,
    cast,
    get_origin,
    runtime_checkable,
)

from fastapi import Request, WebSocket

from jararaca.core.providers import ProviderSpec, T, Token
from jararaca.messagebus import MessageOf
from jararaca.messagebus.types import Message

if TYPE_CHECKING:
    from typing_extensions import TypeIs


@dataclass
class SchedulerAppContext:
    triggered_at: datetime
    scheduled_to: datetime
    cron_expression: str
    action: Callable[..., Any]
    context_type: Literal["scheduler"] = "scheduler"


@dataclass
class HttpAppContext:
    request: Request
    context_type: Literal["http"] = "http"


@dataclass
class MessageBusAppContext:
    topic: str
    message: MessageOf[Message]
    context_type: Literal["message_bus"] = "message_bus"


@dataclass
class WebSocketAppContext:
    websocket: WebSocket
    context_type: Literal["websocket"] = "websocket"


AppContext = (
    MessageBusAppContext | HttpAppContext | SchedulerAppContext | WebSocketAppContext
)

app_context_ctxvar = ContextVar[AppContext]("app_context")


def use_app_context() -> AppContext:
    return app_context_ctxvar.get()


@contextmanager
def provide_app_context(app_context: AppContext) -> Generator[None, None, None]:
    token = app_context_ctxvar.set(app_context)
    try:
        yield
    finally:
        with suppress(ValueError):
            app_context_ctxvar.reset(token)


@runtime_checkable
class AppInterceptor(Protocol):

    def intercept(self, app_context: AppContext) -> AsyncContextManager[None]: ...


class AppInterceptorWithLifecycle(Protocol):

    def lifecycle(
        self, app: "Microservice", container: "Container"
    ) -> AsyncContextManager[None]: ...


def is_interceptor_with_lifecycle(
    interceptor: Any,
) -> "TypeIs[AppInterceptorWithLifecycle]":

    return hasattr(interceptor, "lifecycle")


@dataclass
class Microservice:
    providers: list[type[Any] | ProviderSpec] = field(default_factory=list)
    controllers: list[type] = field(default_factory=list)
    interceptors: list[AppInterceptor | Callable[..., AppInterceptor]] = field(
        default_factory=list
    )


class Container:

    def __init__(self, app: Microservice) -> None:
        self.instances_map: dict[Any, Any] = {}
        self.app = app

    def fill_providers(self, after_interceptors: bool) -> None:
        for provider in (
            app
            for app in self.app.providers
            if after_interceptors == app.after_interceptors
        ):
            if isinstance(provider, ProviderSpec):
                if provider.use_value:
                    self.instances_map[provider.provide] = provider.use_value
                elif provider.use_class:
                    self.get_and_register(provider.use_class, provider.provide)
                elif provider.use_factory:
                    self.get_and_register(provider.use_factory, provider.provide)
            else:
                self.get_and_register(provider, provider)

    def instantiate(self, type_: type[Any] | Callable[..., Any]) -> Any:

        dependencies = self.parse_dependencies(type_)

        evaluated_dependencies = {
            name: self.get_or_register_token_or_type(dependency)
            for name, dependency in dependencies.items()
        }

        instance = type_(**evaluated_dependencies)

        return instance

    def parse_dependencies(
        self, provider: type[Any] | Callable[..., Any]
    ) -> dict[str, type[Any]]:

        signature = inspect.signature(provider)

        parameters = signature.parameters

        return {
            name: self.lookup_parameter_type(parameter)
            for name, parameter in parameters.items()
            if parameter.annotation != inspect.Parameter.empty
        }

    def lookup_parameter_type(self, parameter: inspect.Parameter) -> Any:
        if parameter.annotation == inspect.Parameter.empty:
            raise Exception(f"Parameter {parameter.name} has no type annotation")

        # if is a Annotated type, return the second argument

        annotation = parameter.annotation
        if get_origin(annotation) is Annotated:

            if len(annotation.__metadata__) == 0:
                raise Exception(f"Parameter {parameter.name} has no type annotation")

            return annotation.__metadata__[0]

        return parameter.annotation

    def get_or_register_token_or_type(
        self, token_or_type: Type[T] | Token[T] | Callable[..., T]
    ) -> T:

        if token_or_type in self.instances_map:
            return cast(T, self.instances_map[token_or_type])

        item_type: Type[T] | Callable[..., T]
        bind_to: Token[T] | Type[T] | Callable[..., T]

        if isinstance(token_or_type, Token):
            item_type = token_or_type.type_
            bind_to = token_or_type
        else:
            item_type = bind_to = token_or_type

        if token_or_type not in self.instances_map:
            return self.get_and_register(item_type, bind_to)

        return cast(T, self.instances_map[bind_to])

    def get_and_register(
        self, item_type: Type[T] | Callable[..., T], bind_to: Any
    ) -> T:
        instance = self.instantiate(item_type)
        self.register(instance, bind_to)
        return cast(T, instance)

    def register(self, instance: T, bind_to: Any) -> None:
        self.instances_map[bind_to] = instance

    def get_by_type(self, token: Type[T]) -> T:
        return self.get_or_register_token_or_type(token)

    def get_by_token(self, token: Token[T]) -> T:
        return self.get_or_register_token_or_type(token)


current_container_ctx = ContextVar[Container]("current_container")


def use_current_container() -> Container:
    return current_container_ctx.get()


@contextmanager
def provide_container(container: Container) -> Generator[None, None, None]:
    token = current_container_ctx.set(container)
    try:
        yield
    finally:
        with suppress(ValueError):
            current_container_ctx.reset(token)


__all__ = [
    "AppContext",
    "AppInterceptor",
    "AppInterceptorWithLifecycle",
    "Container",
    "Microservice",
    "SchedulerAppContext",
    "WebSocketAppContext",
    "app_context_ctxvar",
    "current_container_ctx",
    "provide_app_context",
    "provide_container",
    "use_app_context",
    "use_current_container",
]



from typing import ClassVar, Generic, Literal, Protocol, TypeVar

from pydantic import BaseModel

from jararaca.messagebus.publisher import use_publisher


class Message(BaseModel):
    """
    Base class for messages representing tasks.
    A Task is a message that represents a unit of work to be done.
    It is published to a TaskPublisher and consumed by a TaskHandler, wrapped in TaskData.
    Note: A Task is not an Event.
    """

    MESSAGE_TOPIC: ClassVar[str] = "__unset__"

    MESSAGE_TYPE: ClassVar[Literal["task", "event"]] = "task"

    async def publish(self) -> None:
        task_publisher = use_publisher()
        await task_publisher.publish(self, self.MESSAGE_TOPIC)


INHERITS_MESSAGE_CO = TypeVar("INHERITS_MESSAGE_CO", bound=Message, covariant=True)


class MessageOf(Protocol, Generic[INHERITS_MESSAGE_CO]):

    def payload(self) -> INHERITS_MESSAGE_CO: ...



from importlib import import_module
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from jararaca.messagebus.bus_message_controller import (
        ack,
        nack,
        reject,
        retry,
        retry_later,
        use_bus_message_controller,
    )
    from jararaca.microservice import AppContext, AppInterceptor
    from jararaca.observability.interceptor import ObservabilityInterceptor
    from jararaca.observability.providers.otel import OtelObservabilityProvider
    from jararaca.persistence.sort_filter import (
        FILTER_SORT_ENTITY_ATTR_MAP,
        FilterModel,
        FilterRuleApplier,
        SortFilterRunner,
        SortModel,
        SortRuleApplier,
    )
    from jararaca.presentation.hooks import (
        raises_200_on,
        raises_400_on,
        raises_401_on,
        raises_403_on,
        raises_404_on,
        raises_422_on,
        raises_500_on,
        raises_http_exception_on,
    )
    from jararaca.rpc.http.backends.httpx import HTTPXHttpRPCAsyncBackend
    from jararaca.rpc.http.backends.otel import TracedRequestMiddleware
    from jararaca.rpc.http.decorators import Body
    from jararaca.rpc.http.decorators import Delete as HttpDelete
    from jararaca.rpc.http.decorators import Get as HttpGet
    from jararaca.rpc.http.decorators import (
        GlobalHttpErrorHandler,
        Header,
        HttpMapping,
        HttpRpcClientBuilder,
    )
    from jararaca.rpc.http.decorators import Patch as HttpPatch
    from jararaca.rpc.http.decorators import PathParam
    from jararaca.rpc.http.decorators import Post as HttpPost
    from jararaca.rpc.http.decorators import Put as HttpPut
    from jararaca.rpc.http.decorators import (
        Query,
        RequestAttribute,
        RestClient,
        RouteHttpErrorHandler,
        RPCRequestNetworkError,
    )

    from .core.providers import ProviderSpec, Token
    from .di import Container
    from .messagebus.decorators import MessageBusController, MessageHandler
    from .messagebus.interceptors.aiopika_publisher_interceptor import (
        AIOPikaConnectionFactory,
        MessageBusPublisherInterceptor,
    )
    from .messagebus.publisher import use_publisher
    from .messagebus.types import Message, MessageOf
    from .messagebus.worker import MessageBusWorker
    from .microservice import Microservice, use_app_context, use_current_container
    from .persistence.base import T_BASEMODEL, BaseEntity
    from .persistence.interceptors.aiosqa_interceptor import (
        AIOSQAConfig,
        AIOSqlAlchemySessionInterceptor,
        use_session,
    )
    from .persistence.utilities import (
        CriteriaBasedAttributeQueryInjector,
        CRUDOperations,
        DateCriteria,
        DatedEntity,
        DateOrderedFilter,
        DateOrderedQueryInjector,
        Identifiable,
        IdentifiableEntity,
        Paginated,
        PaginatedFilter,
        QueryInjector,
        QueryOperations,
        StringCriteria,
    )
    from .presentation.decorators import (
        Delete,
        Get,
        Patch,
        Post,
        Put,
        RestController,
        UseDependency,
        UseMiddleware,
    )
    from .presentation.http_microservice import HttpMicroservice, HttpMiddleware
    from .presentation.server import create_http_server
    from .presentation.websocket.context import (
        WebSocketConnectionManager,
        provide_ws_manager,
        use_ws_manager,
    )
    from .presentation.websocket.decorators import (
        RegisterWebSocketMessage,
        WebSocketEndpoint,
    )
    from .presentation.websocket.redis import RedisWebSocketConnectionBackend
    from .presentation.websocket.types import WebSocketMessage
    from .presentation.websocket.websocket_interceptor import WebSocketInterceptor
    from .scheduler.decorators import ScheduledAction
    from .tools.app_config.interceptor import AppConfigurationInterceptor

    __all__ = [
        "FilterRuleApplier",
        "SortRuleApplier",
        "use_bus_message_controller",
        "ack",
        "nack",
        "reject",
        "retry",
        "retry_later",
        "RPCRequestNetworkError",
        "FILTER_SORT_ENTITY_ATTR_MAP",
        "FilterModel",
        "SortFilterRunner",
        "SortModel",
        "RegisterWebSocketMessage",
        "TracedRequestMiddleware",
        "raises_http_exception_on",
        "raises_200_on",
        "raises_422_on",
        "raises_404_on",
        "raises_400_on",
        "raises_401_on",
        "raises_403_on",
        "raises_500_on",
        "HttpMiddleware",
        "HttpMapping",
        "RequestAttribute",
        "Body",
        "Query",
        "Header",
        "PathParam",
        "RestClient",
        "HttpPost",
        "HttpGet",
        "HttpPatch",
        "HttpPut",
        "HttpDelete",
        "ObservabilityInterceptor",
        "QueryInjector",
        "HttpMicroservice",
        "use_current_container",
        "T_BASEMODEL",
        "DatedEntity",
        "BaseEntity",
        "use_ws_manager",
        "WebSocketEndpoint",
        "CriteriaBasedAttributeQueryInjector",
        "Identifiable",
        "IdentifiableEntity",
        "MessageOf",
        "Message",
        "StringCriteria",
        "DateCriteria",
        "DateOrderedFilter",
        "DateOrderedQueryInjector",
        "Paginated",
        "PaginatedFilter",
        "QueryOperations",
        "CRUDOperations",
        "RestController",
        "MessageBusController",
        "MessageHandler",
        "ScheduledAction",
        "Microservice",
        "ProviderSpec",
        "Token",
        "AIOSqlAlchemySessionInterceptor",
        "AIOSQAConfig",
        "create_http_server",
        "MessageBusWorker",
        "Container",
        "WebSocketInterceptor",
        "use_session",
        "Post",
        "Get",
        "Patch",
        "Put",
        "Delete",
        "use_publisher",
        "AIOPikaConnectionFactory",
        "MessageBusPublisherInterceptor",
        "RedisWebSocketConnectionBackend",
        "AppConfigurationInterceptor",
        "UseMiddleware",
        "UseDependency",
        "GlobalHttpErrorHandler",
        "RouteHttpErrorHandler",
        "WebSocketMessage",
        "WebSocketConnectionManager",
        "provide_ws_manager",
        "HttpRpcClientBuilder",
        "HTTPXHttpRPCAsyncBackend",
        "use_app_context",
        "AppContext",
        "AppInterceptor",
        "OtelObservabilityProvider",
    ]

__SPEC_PARENT__: str = __spec__.parent  # type: ignore
# A mapping of {<member name>: (package, <module name>)} defining dynamic imports
_dynamic_imports: "dict[str, tuple[str, str, str | None]]" = {
    "FilterRuleApplier": (__SPEC_PARENT__, "persistence.sort_filter", None),
    "SortRuleApplier": (__SPEC_PARENT__, "persistence.sort_filter", None),
    "use_bus_message_controller": (
        __SPEC_PARENT__,
        "messagebus.bus_message_controller",
        None,
    ),
    "ack": (__SPEC_PARENT__, "messagebus.bus_message_controller", None),
    "nack": (__SPEC_PARENT__, "messagebus.bus_message_controller", None),
    "reject": (__SPEC_PARENT__, "messagebus.bus_message_controller", None),
    "retry": (__SPEC_PARENT__, "messagebus.bus_message_controller", None),
    "retry_later": (__SPEC_PARENT__, "messagebus.bus_message_controller", None),
    "RPCRequestNetworkError": (__SPEC_PARENT__, "rpc.http.decorators", None),
    "FILTER_SORT_ENTITY_ATTR_MAP": (__SPEC_PARENT__, "persistence.sort_filter", None),
    "FilterModel": (__SPEC_PARENT__, "persistence.sort_filter", None),
    "SortFilterRunner": (__SPEC_PARENT__, "persistence.sort_filter", None),
    "SortModel": (__SPEC_PARENT__, "persistence.sort_filter", None),
    "RegisterWebSocketMessage": (
        __SPEC_PARENT__,
        "presentation.websocket.decorators",
        None,
    ),
    "OtelObservabilityProvider": (
        __SPEC_PARENT__,
        "observability.providers.otel",
        None,
    ),
    "TracedRequestMiddleware": (__SPEC_PARENT__, "rpc.http.backends.otel", None),
    "raises_http_exception_on": (__SPEC_PARENT__, "presentation.hooks", None),
    "raises_200_on": (__SPEC_PARENT__, "presentation.hooks", None),
    "raises_400_on": (__SPEC_PARENT__, "presentation.hooks", None),
    "raises_401_on": (__SPEC_PARENT__, "presentation.hooks", None),
    "raises_403_on": (__SPEC_PARENT__, "presentation.hooks", None),
    "raises_404_on": (__SPEC_PARENT__, "presentation.hooks", None),
    "raises_422_on": (__SPEC_PARENT__, "presentation.hooks", None),
    "raises_500_on": (__SPEC_PARENT__, "presentation.hooks", None),
    "HttpMiddleware": (__SPEC_PARENT__, "presentation.http_microservice", None),
    "HttpMapping": (__SPEC_PARENT__, "rpc.http.decorators", None),
    "RequestAttribute": (__SPEC_PARENT__, "rpc.http.decorators", None),
    "Body": (__SPEC_PARENT__, "rpc.http.decorators", None),
    "Query": (__SPEC_PARENT__, "rpc.http.decorators", None),
    "Header": (__SPEC_PARENT__, "rpc.http.decorators", None),
    "PathParam": (__SPEC_PARENT__, "rpc.http.decorators", None),
    "RestClient": (__SPEC_PARENT__, "rpc.http.decorators", None),
    "HttpPost": (__SPEC_PARENT__, "rpc.http.decorators", "Post"),
    "HttpGet": (__SPEC_PARENT__, "rpc.http.decorators", "Get"),
    "HttpPatch": (__SPEC_PARENT__, "rpc.http.decorators", "Patch"),
    "HttpPut": (__SPEC_PARENT__, "rpc.http.decorators", "Put"),
    "HttpDelete": (__SPEC_PARENT__, "rpc.http.decorators", "Delete"),
    "ObservabilityInterceptor": (__SPEC_PARENT__, "observability.interceptor", None),
    "QueryInjector": (__SPEC_PARENT__, "persistence.utilities", None),
    "HttpMicroservice": (__SPEC_PARENT__, "presentation.http_microservice", None),
    "use_current_container": (__SPEC_PARENT__, "microservice", None),
    "T_BASEMODEL": (__SPEC_PARENT__, "persistence.base", None),
    "DatedEntity": (__SPEC_PARENT__, "persistence.utilities", None),
    "BaseEntity": (__SPEC_PARENT__, "persistence.base", None),
    "use_ws_manager": (__SPEC_PARENT__, "presentation.websocket.context", None),
    "WebSocketEndpoint": (__SPEC_PARENT__, "presentation.websocket.decorators", None),
    "CriteriaBasedAttributeQueryInjector": (
        __SPEC_PARENT__,
        "persistence.utilities",
        None,
    ),
    "Identifiable": (__SPEC_PARENT__, "persistence.utilities", None),
    "IdentifiableEntity": (__SPEC_PARENT__, "persistence.utilities", None),
    "MessageOf": (__SPEC_PARENT__, "messagebus.types", None),
    "Message": (__SPEC_PARENT__, "messagebus.types", None),
    "StringCriteria": (__SPEC_PARENT__, "persistence.utilities", None),
    "DateCriteria": (__SPEC_PARENT__, "persistence.utilities", None),
    "DateOrderedFilter": (__SPEC_PARENT__, "persistence.utilities", None),
    "DateOrderedQueryInjector": (__SPEC_PARENT__, "persistence.utilities", None),
    "Paginated": (__SPEC_PARENT__, "persistence.utilities", None),
    "PaginatedFilter": (__SPEC_PARENT__, "persistence.utilities", None),
    "QueryOperations": (__SPEC_PARENT__, "persistence.utilities", None),
    "CRUDOperations": (__SPEC_PARENT__, "persistence.utilities", None),
    "RestController": (__SPEC_PARENT__, "presentation.decorators", None),
    "MessageBusController": (__SPEC_PARENT__, "messagebus.decorators", None),
    "MessageHandler": (__SPEC_PARENT__, "messagebus.decorators", None),
    "ScheduledAction": (__SPEC_PARENT__, "scheduler.decorators", None),
    "Microservice": (__SPEC_PARENT__, "microservice", None),
    "ProviderSpec": (__SPEC_PARENT__, "core.providers", None),
    "Token": (__SPEC_PARENT__, "core.providers", None),
    "AIOSqlAlchemySessionInterceptor": (
        __SPEC_PARENT__,
        "persistence.interceptors.aiosqa_interceptor",
        None,
    ),
    "AIOSQAConfig": (
        __SPEC_PARENT__,
        "persistence.interceptors.aiosqa_interceptor",
        None,
    ),
    "create_http_server": (__SPEC_PARENT__, "presentation.server", None),
    "MessageBusWorker": (__SPEC_PARENT__, "messagebus.worker", None),
    "Container": (__SPEC_PARENT__, "di", None),
    "WebSocketInterceptor": (
        __SPEC_PARENT__,
        "presentation.websocket.websocket_interceptor",
        None,
    ),
    "use_session": (
        __SPEC_PARENT__,
        "persistence.interceptors.aiosqa_interceptor",
        None,
    ),
    "Post": (__SPEC_PARENT__, "presentation.decorators", None),
    "Get": (__SPEC_PARENT__, "presentation.decorators", None),
    "Patch": (__SPEC_PARENT__, "presentation.decorators", None),
    "Put": (__SPEC_PARENT__, "presentation.decorators", None),
    "Delete": (__SPEC_PARENT__, "presentation.decorators", None),
    "use_publisher": (__SPEC_PARENT__, "messagebus.publisher", None),
    "AIOPikaConnectionFactory": (
        __SPEC_PARENT__,
        "messagebus.interceptors.aiopika_publisher_interceptor",
        None,
    ),
    "MessageBusPublisherInterceptor": (
        __SPEC_PARENT__,
        "messagebus.interceptors.aiopika_publisher_interceptor",
        None,
    ),
    "RedisWebSocketConnectionBackend": (
        __SPEC_PARENT__,
        "presentation.websocket.redis",
        None,
    ),
    "AppConfigurationInterceptor": (
        __SPEC_PARENT__,
        "tools.app_config.interceptor",
        None,
    ),
    "UseMiddleware": (__SPEC_PARENT__, "presentation.decorators", None),
    "UseDependency": (__SPEC_PARENT__, "presentation.decorators", None),
    "GlobalHttpErrorHandler": (__SPEC_PARENT__, "rpc.http.decorators", None),
    "RouteHttpErrorHandler": (__SPEC_PARENT__, "rpc.http.decorators", None),
    "WebSocketMessage": (__SPEC_PARENT__, "presentation.websocket.types", None),
    "WebSocketConnectionManager": (
        __SPEC_PARENT__,
        "presentation.websocket.context",
        None,
    ),
    "provide_ws_manager": (__SPEC_PARENT__, "presentation.websocket.context", None),
    "HttpRpcClientBuilder": (__SPEC_PARENT__, "rpc.http.decorators", None),
    "HTTPXHttpRPCAsyncBackend": (__SPEC_PARENT__, "rpc.http.backends.httpx", None),
    "use_app_context": (__SPEC_PARENT__, "microservice", None),
    "AppContext": (__SPEC_PARENT__, "microservice", None),
    "AppInterceptor": (__SPEC_PARENT__, "microservice", None),
}


def __getattr__(attr_name: str) -> object:

    dynamic_attr = _dynamic_imports.get(attr_name)
    if dynamic_attr is None:
        raise AttributeError(f"module {__name__!r} has no attribute {attr_name!r}")

    package, module_name, realname = dynamic_attr

    if module_name == "__module__":
        result = import_module(f".{attr_name}", package=package)
        globals()[attr_name] = result
        return result
    else:
        module = import_module(f"{package}.{module_name}", package=package)
        result = getattr(module, attr_name if realname is None else realname)
        g = globals()
        g[attr_name] = result
        for k, (_, _, realname) in _dynamic_imports.items():
            # if v_module_name == module_name and k not in _deprecated_dynamic_imports:
            g[k] = getattr(module, k if realname is None else realname)
        return result


def __dir__() -> "list[str]":
    return list(__all__)
